'EVE2 driver for Newhaven LCD display modules by Raymond Allen  
'MIT license

OBJ
  SPI:  "SPI_Asm_RJA2"  '"SPI_Spin"'   'You can either use Assembly or Spin (Spin is slower, but saves a cog)
  EEPROM: "Propeller EEPROM" 'Used to save resistive touchscreen calibrations to eeprom

VAR
  long  pin_sck, pin_miso, pin_mosi, pin_csn 'SPI interface pins
  long  pin_intn, pin_pdn 'EVE interrupt and power down pins
  long  CommandPointer  'Points to current position in command buffer.  This gets added to RAM_CMD
  long  DisplayListPointer 'Pointer to current position in display list (for next direct write to display list)
  long  tracker 'Data returned from REG_TRACKER, needed to return extended info
  'Display settings:
  long  DispWidth,DispHeight,DispHCycle,DispHOffset,DispHSync0,DispHSync1,DispVCycle,DispVOffset,DispVSync0,DispVSync1,DispPCLK,DispSwizzle,DispPCLKPol,DispCSpread,DispDither
  
CON     InitBrightness=128/2  'Initial display brightness 128=max, 0=off



PUB Start(pSettings,sck,miso,mosi,csn,intn,pdn)|i,j 'transfer settings, initialize SPI interface and activate EVE given all pins
    'Bring in display settings
    DispWidth := long[pSettings][0]
    DispHeight := long[pSettings][1] 
    DispHCycle :=  long[pSettings][2] 
    DispHOffset := long[pSettings][3] 
    DispHSync0 := long[pSettings][4] 
    DispHSync1 := long[pSettings][5] 
    DispVCycle := long[pSettings][6] 
    DispVOffset := long[pSettings][7] 
    DispVSync0 := long[pSettings][8] 
    DispVSync1 := long[pSettings][9] 
    DispPCLK := long[pSettings][10] 
    DispSwizzle := long[pSettings][11] 
    DispPCLKPol := long[pSettings][12] 
    DispCSpread := long[pSettings][13] 
    DispDither := long[pSettings][14] 

    'Assing pins to those given
     pin_sck:=sck
     pin_miso:=miso
     pin_mosi:=mosi
     pin_csn:=csn
     pin_intn:=intn
     pin_pdn:=pdn
      
    'We will cycle the power down pin to force reset of EVE2
    high(pin_csn)
    low(pin_pdn)
    waitcnt(cnt+clkfreq/10)
    high(pin_pdn)
    waitcnt(cnt+clkfreq/10)

    '' Initialize SPI Engine with Clock Delay of 15 and Clock State of 0 
    SPI.start(10,0)             

    'Send Active command
    low(pin_csn)
    SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 24, 0)   'send address + dummy byte
    high(pin_csn)
    low(pin_csn)
    SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 24, 0)   'send address + dummy byte
    high(pin_csn)
    
    'Wait until ID reads properly
    repeat          
      i:=$302000  'ID_Reg
      j:=ReadReg8(i)
      waitcnt(cnt+clkfreq/10)
    until j==$7C

    'Set display vaules
    wr16(REG_HCYCLE, DispHCycle)
    wr16(REG_HOFFSET, DispHOffset)
    wr16(REG_HSYNC0, DispHSync0)
    wr16(REG_HSYNC1, DispHSync1)
    wr16(REG_VCYCLE, DispVCycle)
    wr16(REG_VOFFSET, DispVOffset)
    wr16(REG_VSYNC0, DispVSync0)
    wr16(REG_VSYNC1, DispVSync1)
    wr8(REG_SWIZZLE, DispSwizzle)
    wr8(REG_PCLK_POL, DispPCLKPol)
    wr8(REG_CSPREAD, DispCSpread)
    wr16(REG_HSIZE, DispWidth)
    wr16(REG_VSIZE, DispHeight)

    'Adjust brightness
    wr8(REG_PWM_DUTY, InitBrightness)  '128=max, 0=off

    'Enable display bit
    WR8(REG_GPIO_DIR,RD8(REG_GPIO_DIR)|$80)
    WR8(REG_GPIO,RD8(REG_GPIO)|$80)

    'Set pixel clock
    WR8(REG_PCLK,DispPCLK)

    'Initialize co-processor commmand pointer
    CommandPointer:=0  

PUB SetBrightness(Brightness) 'set LCD backlight brightness:  128=max, 0=off
  wr8(REG_PWM_DUTY, Brightness)  '128=max, 0=off

PUB StartDisplayList 'reset display list pointer to beginning of display list
  DisplayListPointer:=RAM_DL

PUB ClearScreen(R,G,B)  'Clear the screen to specified color   
  WrDl32(Clear_Color_RGB(r,g,b)) 'set color
  WrDl32(CLEAR(1, 1, 1)) 'clear everything

  
PUB EndDisplayList  'send completed display list to screen
  WrDl32(DISPLAY)  'mark end of display list
  wr8(REG_DLSWAP,%10) 'update after current frame

PUB DrawCharacter(x, y, font, char) 'Draw character char using font #font to x,y location (x,y limited to 0..511)
  'can also be used to draw bitmaps
  WrDl32(VERTEX2II(x, y, font, char))

PUB BeginDrawingBitmaps  'Start list of bitmaps to draw
  WrDl32(BeginBitmaps)  'start drawing bitmaps

PUB CmdBeginBitmaps
  CMD(BeginBitmaps)
    
  
PUB EndBegin  'mark end of "Begin" section
  WrDl32(End)  'end drawing bitmaps

PUB CmdEndBegin
  CMD(End)
  
PUB SetColor(r, g, b) 'Set color for subsequent operations to RGB(r,g,b)
  WrDl32(COLOR_RGB(r, g, b))

PUB SetPointSize(sz) 'set point size to 20 pixels in radius (units of 1/16 pixel, 20*16=320)         
  WrDl32(POINT_SIZE(sz<<4))

PUB TranslateX(x)   'shift subsequent points by x/16 pixels in x direction
  WrDl32(Vertex_Translate_X(x<<4))

PUB TranslateY(y)   'shift subsequent points by y/16 pixels in x direction
  WrDl32(Vertex_Translate_Y(y<<4))

PUB BeginDrawingPoints  'Start list of points to draw
  WrDl32(VertexFormat(4)) 'Select vertex format 4 (1/16 pixel precision, range -1024 to 1023) 
  WrDl32(BEGINPOINTS)

PUB DrawPoint(x,y)  'draw a point at x,y in units of 1/16 pixel
  WrDl32(VERTEX2F(x<<4, y<<4))  
    
PUB StartCmdDisplayList 'Start a new display list using the Co-Processor
  CMD(CMD_DLSTART)

PUB CmdClearColor(r,g,b) 'set clear screen color to RGB(r,g,b)
  CMD(CLEAR_COLOR_RGB(r, g, b))

PUB CmdClearScreen 'clear screen to screen clear color
  CMD(Clear(1,1,1))

PUB CmdButton(x,y,w,h,font,options,pString) 'draw button at x,y with dimensions w,h using font# font, options and pointer to string
  button(x,y,w,h,font,options,pString)     

PUB CmdClock(x,y,r,options,h,m,s,ms) 'draw a clock at x,y with radius of r and options at time set to h hours, m minutes, s seconds, ms milliseconds
  Clock(x,y,r,options,h,m,s,ms)

PUB CmdDial(x,y,r,options,val) 'draw a rotary knob , value is from 0 to 65535 with 0 pointing down
  CMD(CMD_Dial)
  CMD(Y<<16+X)
  CMD(options<<16+r)
  CMD(val)

PUB CmdSlider(x,y,w,h,options,val,range) 'draw a horizontal slider control
  CMD(CMD_Slider)
  CMD(Y<<16+X)
  CMD(h<<16+w)   
  CMD(val<<16+options)
  CMD(range)

PUB CmdLoadEmbeddedImage(pImage, nBytes, ptr, options)|d 'Load image from HUB RAM, pImage is location in Ram, size is nbytes, ptr is pointer to gram where image is to go, options
  CMD(CMD_LoadImage)
  CMD(ptr)
  CMD(options)
  repeat (nBytes+3)/4
    d:=byte[pImage][3]<<24
    d+=byte[pImage][2]<<16
    d+=byte[pImage][1]<<8
    d+=byte[pImage][0]
    CMD(d)
    pImage+=4

PUB CmdBitmapSource(pSource) 'specify start address of bitmap in gram
  CMD($01<<24+pSource)

PUB CmdBitmapLayoutRGB565(width,height)|linestride 'set layout to RGB565 with provided width and height
  linestride:=width*2 '2 bytes per pixel for RGB565
  CMD($28<<24+(linestride>>10)<<2) 'upper 2 bits of linestride
  CMD($07<<24+7<<19+(linestride&$3FF)<<9+height) '7=RGB565  , lower 10 bits of linestride and height      

PUB CmdBitmapSize(w,h) 'width, height  'for >511 need to set upper bits
  CMD($08<<24+(w&$1FF)<<9+h) 'Bitmap_Size
  CMD($29<<24+(w>>9)<<2+h>>9)'Bitmap_Size_H 'upper bits of bitmap size

PUB CmdVertex2F(x,y)
  CMD(Vertex2F(x,y))

PUB ZeroMemory(p,n) 'zero n bytes of memory starting at p
  CMD(CMD_MemZero)
  CMD(p)
  CMD(n)

PUB EndCmdDisplayList 'End and show display list using the Co-Processor
  CMD(DISPLAY)
  CMD(CMD_DLSWAP)

PUB CmdGuage(x,y,r,options,major,minor,val,range)  'draw an analog guage at x,y with radius r, # of major and minor ticks, value and range
  Gauge(x,y,r,options,major,minor,val,range)      
    
PUB CmdRomFont(font,romslot)'Assign font# font to slot# romslot (needed only to use the biggest fonts, >#31) 
  ROMFONT(font,romslot)

PUB CmdText(x,y,font,options,pString) 'Draw text at x,y using font# font, options and provided string pointer   
  Text(x,y,font,options,pString)

PUB bCapTagTouched(tag):b 'Capacitive Touch:  Returns true if tag was touched, false otherwise
  tracker:=ReadReg32(REG_CTOUCH_TAG)
  if ((tracker&$FF) == tag)
    return tracker
  return false

PUB WaitIdle
  repeat
      waitcnt(cnt+clkfreq/100)
  until ReadReg32(REG_CMD_READ) == ReadReg32(REG_CMD_WRITE) 

PUB bResTagTouched(tag):b 'Resistive Touch:  Returns true if tag was touched, false otherwise
  'return ReadReg32(REG_TOUCH_SCREEN_XY)
  tracker:=ReadReg32(REG_TOUCH_TAG)
  return tracker
  if ((tracker&$FF) == tag)
    return tag
  return false

PUB TrackTouch(x,y,w,h,tag) 'Track touch for area with corner at (x,y), width w, height h and give it tag for tracking
  CMD(CMD_TRACK)
  CMD(y<<16+x)
  CMD(h<<16+w)
  CMD(tag)

PUB GetTrack  'return contents of tracker register
  return ReadReg32(REG_TRACKER)

PUB EnableTagging   'allow tagging of objects for touch detection
  CMD($14<<24+1)  '($14<<24+1)

PUB RestoreEepromCals(pSource,pDest)|i,j 'move cal factors from upper eeprom to lower eeprom
  'cal factors are in upper EEPROM, copy to dest in RAM
  EEPROM.ToRam(pDest,pDest+4*7+3,pSource)

PUB SaveEepromCals(pSource,pDest)
  EEPROM.FromRam(pSource,pSource+4*7+3,pDest)  


PUB RestoreResistiveCalibration(pFactors)|i 'Send resistive touchscreen calibration factors to EVE
  'cal factors were in lower eeprom and restored when booted from eeprom
  repeat i from 0 to 5
    WriteReg32(REG_TOUCH_TRANSFORM_A+i<<2,long[pFactors][i])

PUB GetCalFactors(pFactors)|i
  repeat i from 0 to 5
    long[pfactors][i]:=ReadReg32(REG_TOUCH_TRANSFORM_A+i<<2)
{
PUB SaveResistiveCalibration(pCode) 'Save calibrated code and resistive cal factors to eeprom so don't always need to do calibration
  long[pCode]:=$7C 'set code to indicate that cal factors are saved
  EEPROM.VarBackup(pCode+4,pCode+4*7+3)
}      

PUB CmdSetBase(base) 'switch from default of decimal
  CMD(CMD_SETBASE)
  CMD(base)

PUB CmdNumber(x,y,font,options,n) 'draw number n at x,y using font and options(or #of digits with one of following:  EVE#OPT_CENTERX, EVE#OPT_CENTERT, EVE#OPT_CENTER, EVE#OPT_RIGHTX, EVE#OPT_SIGNED)
  CMD(CMD_NUMBER)
  CMD(Y<<16+X)
  CMD(options<<16+font)
  CMD(n)

PUB TagFollowing(t) 'tag the next graphics objects with tag=t
  CMD($3<<24+t)

PUB CMD(command)
   'wr32(RAM_CMD+CommandPointer,command)
   'commandPointer+=4
   'CommandPointer&=$FFF  '4k circular buffer
   'wr16(REG_CMD_WRITE,CommandPointer)
   wr32(REG_CMDB_WRITE,command)


   
PUB Button(x,y,w,h,font,options,pString)|i,j
  CMD(CMD_BUTTON)
  CMD(y<<16+x)
  CMD(h<<16+w)
  CMD(options<<16+font)
  j:=(strsize(pString)+4)/4 
  repeat i from 1 to j
    CMD(byte[pString+3]<<24+byte[pString+2]<<16+byte[pString+1]<<8+byte[pString])
    pString+=4

PUB Text(x,y,font,options,pString)|i,j
  CMD(CMD_TEXT)
  CMD(y<<16+x)
  CMD(options<<16+font)
  j:=(strsize(pString)+4)/4 
  repeat i from 1 to j
    CMD(byte[pString+3]<<24+byte[pString+2]<<16+byte[pString+1]<<8+byte[pString])
    pString+=4

PUB Clock(x,y,r,options,h,m,s,ms)
  CMD(CMD_CLOCK)
  CMD(y<<16+x)
  CMD(options<<16+r)
  CMD(m<<16+h)
  CMD(ms<<16+s)

PUB Gauge(x,y,r,options,major,minor,val,range)
  CMD(CMD_GAUGE)
  CMD(y<<16+x)
  CMD(options<<16+r)
  CMD(minor<<16+major)
  CMD(range<<16+val)   

PUB ROMFONT(font,romslot)
  CMD(CMD_ROMFONT)
  CMD(font)
  CMD(romslot)

PUB CmdCalibrate
  CMD(CMD_CALIBRATE)
  CMD(0)
  
PUB Clear_Color_RGB(r,g,b)
  return $02<<24+r<<16+g<<8+b
PUB Color_RGB(r,g,b)
  return $04<<24+r<<16+g<<8+b  
PUB Clear(c,s,t) 'c=1 to clear color buffer, s=1 to clear stencil buffer, t=1 to clear tag buffer
  return $26<<24+(c&1)<<2+(s&1)<<1+(t&1)
PUB Display
  return 0
PUB Vertex2II(x,y,handle,cell)
  return $2<<30+x<<21+y<<12+handle<<7+cell
PUB Vertex2F(x,y)
  return $1<<30+x<<15+y  
PUB Vertex_Translate_X(dx)
  return ($2B<<24+dx)
PUB Vertex_Translate_Y(dy)
  return ($2C<<24+dy)    
PUB BeginBitmaps
  return $1F<<24+1
PUB End
  return $21<<24
PUB Point_Size(size)
  return $0D<<24+size
PUB BeginPoints
  return $1F<<24+2
PUB VertexFormat(frac)
  return $27<<24+frac  



CON  'Co-processor commands
        CMD_DLSTART =  $ffffff00
        CMD_DLSWAP =   $ffffff01
        CMD_BUTTON =   $ffffff0d
        CMD_CLOCK =    $ffffff14
        CMD_GAUGE =    $ffffff13
        CMD_ROMFONT =  $ffffff3f
        CMD_TEXT =     $ffffff0c
        CMD_TRACK    = $ffffff2c
        CMD_SETBASE =  $ffffff38
        CMD_NUMBER =   $ffffff2e 
        CMD_CALIBRATE= $ffffff15
        CMD_Dial =     $ffffff2d
        CMD_Slider =   $ffffff10
        CMD_LoadImage =$ffffff24
        CMD_MemZero =  $ffffff1C
        
        
CON 'EVE Registers 
        'Graphics Registers
        REG_HCYCLE =    $30202C 'Bit 11 - 0: These bits are the number of total PCLK cycles per horizontal line scan
        REG_HOFFSET =   $302030 'Bit 11 - 0:  These bits are used to specify the numbers of PCLK cycles before pixels are scanned out. 
        REG_HSYNC0 =    $302038 'Bit 11 - 0:  The value of these bits specifies how many PCLK cycles of HSYNC high state during start of line. 
        REG_HSYNC1 =    $30203C 'Bit 11 - 0:  The value of these bits specifies how many PCLK cycles for HSYNC during start of line. 
        REG_VCYCLE =    $302040 'Bit 11 - 0:  The value of these bits specifies how many lines in one frame. 
        REG_VOFFSET =   $302044 'Bit 11 - 0:  The value of these bits specifies how many lines taken after the start of new frame.
        REG_VSYNC0 =    $30204C 'Bit 11 - 0:  The value of these bits specifies how many lines for the high state of signal VSYNC  takes at the start of new frame. 
        REG_VSYNC1 =    $302050 'Bit 11 - 0:  The value of these bits specifies how many lines for signal VSYNC takes at the start of new frame. 
        REG_SWIZZLE =   $302064 'Bit 3 - 0 :  These bits are set to control the arrangement of output RGB pins, which may help support different LCD panel.
        REG_PCLK_POL =  $30206C 'Bit 0 :    This bit controls the polarity of PCLK. If it is set to zero, PCLK polarity is on the rising edge.
        REG_CSPREAD=    $302068 'Bit 0 :  This bit controls the transition of RGB signals with PCLK active clock edge, which helps reduce the system noise 
        REG_HSIZE =     $302034 'Bit  11 - 0:  These bits are used to specify the numbers of PCLK cycles per horizonal line.
        REG_VSIZE =     $302048 'Bit 11 - 0:  The value of these bits specifies how many lines of pixels in one frame.  The valid range is from 0 to 2048
        REG_PCLK =      $302070  'Bit 7 - 0 :   These bits  are set to divide the main clock for PCLK. 
        REG_DLSWAP =    $302054  'Bit 1 - 0:   These bits can be set by the host to validate the display list buffer
        REG_ROTATE =    $302058  'Bit 2~0:   screen rotation control bits.
        REG_OUTBITS =   $30205C 'Bit 8 - 0:  These 9 bits are split into 3 groups for Red, Green and Blue color output signals:
        REG_DITHER =    $302060  'Bit 0 :  Set to 1 to enable dithering feature on  RGB signals output. Set to 0 to disable dithering feature.             
        'Graphics Touch registers
        REG_TAG =       $30207C 'Bit 7 - 0 :  These bits are  updated with tag value by FT81X graphics engine.  The tag value here is corresponding to the touching point coordinator given in REG_TAG_X and REG_TAG_Y.
        REG_TAG_Y =     $302078 'Bit 10 - 0 :  These bits are set by the host as Y coordinate of touching point, which will enable the host to query the tag value. This register shall be used together with REG_TAG_X and REG_TAG.
        REG_TAG_X =     $302074 'Bit 10 - 0 :  These bits are set by the host as X coordinate of touching point, which will enable host to query the tag value. This register shall be used together with REG_TAG_Y and REG_TAG
        'Audio registers
        REG_PLAY =      $30208C 'Bit 0 :  A write to this bit triggers the play of the synthesized sound effect specified in REG_SOUND.
        REG_SOUND =     $302088 'Bit 0 - 15 :  These bits are used to select the synthesized sound effect.  They are split into two group Bit 7 - 0, Bit 15 - 8.
        REG_VOL_SOUND = $302084 'Bit 7 -  0:  These bits control the volume of the synthesizer sound.  The default value 0xFF is highest volume. The value zero means mute. 
        REG_VOL_PB    = $302080 'Bit 7 - 0 :  These bits control the volume of the audio file playback.  The default value 0xFF is highest volume. The value zero means mute.
        'Touch Engine Registers
        REG_TOUCH_CONFIG =      $302168 'Configuration register
        REG_TOUCH_TRANSFORM_F = $302164 'Transform coefficient  matrix
        REG_TOUCH_TRANSFORM_E = $302160 'Transform coefficient  matrix
        REG_TOUCH_TRANSFORM_D = $30215C 'Transform coefficient  matrix
        REG_TOUCH_TRANSFORM_C = $302158 'Transform coefficient  matrix
        REG_TOUCH_TRANSFORM_B = $302154 'Transform coefficient  matrix
        REG_TOUCH_TRANSFORM_A = $302150 'Transform coefficient  matrix
        'Resistive Touch engine
        REG_TOUCH_MODE =        $302104 ' Touch screen sampling Mode 
        REG_TOUCH_ADC_MODE =    $302108 ' Select ADC working mode  
        REG_TOUCH_CHARGE =      $30210C' Touch screen charge time, unit of 6 clocks 
        REG_TOUCH_SETTLE =      $302110'Touch screen settle time, unit of 6 clocks 
        REG_TOUCH_OVERSAMPLE =  $302114' Touch screen oversample factor 
        REG_TOUCH_RZTHRESH =    $302118'Touch screen resistance threshold 
        REG_TOUCH_RAW_XY =      $30211C'Touch screen raw x,y(16,16) 
        REG_TOUCH_RZ =          $302120 ' Touch screen resistance 
        REG_TOUCH_SCREEN_XY =   $302124' Touch screen x,y(16,16) 
        REG_TOUCH_TAG =         $30212C' Touch screen Tag result
        'Capacitive Touch Engine
        REG_CTOUCH_MODE =       $302104      ' Touch screen sampling Mode 
        REG_CTOUCH_EXTENDED =   $302108  'Select ADC working mode  
        REG_CTOUCH_TOUCH1_XY =  $30211C 'Coordinate of second touch point 
        REG_CTOUCH_TOUCH4_Y =   $302120  '  Y coordinate of fifth touch point  
        REG_CTOUCH_TOUCH_XY =   $302124  ' Coordinate of first touch point 
        REG_CTOUCH_TAG_XY =     $302128    ' coordinate used to calculate the tag of first touch point 
        REG_CTOUCH_TAG =        $30212C       ' Touch screen Tag result of fist touch point 
        REG_CTOUCH_TAG1_XY =    $302130   ' XY used to tag of second touch point 
        REG_CTOUCH_TAG1 =       $302134      'Tag result of second touch point 
        REG_CTOUCH_TAG2_XY =    $302138   ' XY used to tag of third  touch point 
        REG_CTOUCH_TAG2 =       $30213C      'Tag result of third touch point 
        REG_CTOUCH_TAG3_XY =    $302140   ' XY used to tag of fourth               
        REG_CTOUCH_TAG3 =       $302144       ' Tag result of fourth touch point 
        REG_CTOUCH_TAG4_XY =    $302148   ' XY used to tag of fifth  touch point 
        REG_CTOUCH_TAG4 =       $30214C      ' Tag result of fifth touch point 
        REG_CTOUCH_TOUCH4_X =   $30216C  ' X coordinate of fifth touch point 
        REG_CTOUCH_TOUCH2_XY =  $30218C ' Third touch point coordinate 
        REG_CTOUCH_TOUCH3_XY =  $302190 ' Fourth touch point coordinate
        'Co-processor Engine Registers
        REG_CMD_DL =            $302100 'Bit 12 - 0 :  These bits indicate the offset from RAM_DL of a display list command generated by the coprocessor engine.
        REG_CMD_WRITE =         $3020FC 'Bit 11 - 0 :  These bits are updated by the host to inform the co-processor engine of the ending address of valid data feeding into its FIFO.
        REG_CMD_READ =          $3020F8 'Bit 11 - 0:  These bits are updated by the coprocessor engine as long as the coprocessor engine fetched the command from its FIFO
        REG_CMDB_SPACE =        $302574 'Bit 11 - 0 :  These bits are updated by the coprocessor engine to indicate the free space in RAM_CMD.
        REG_CMDB_WRITE =        $302578 'Bit 31 - 0 :  The data or command to be written into RAM_CMD
        'Special Registers, not in RAM_REG
        REG_TRACKER =           $309000
        REG_TRACKER_1 =         $309004
        REG_TRACKER_2 =         $309008
        REG_TRACKER_3 =         $30900C
        REG_TRACKER_4 =         $309010
        REG_MEDIAFIFO_READ =    $309014 'Bit 31 - 0:  The value specifies the read pointer pointing to the address in RAM_G of the media FIFO
        REG_MEDIAFIFO_WRITE =   $309018 'Bit 31 - 0:  The value specifies the write pointer pointing to the address in RAM_G of the media FIFO.
        'Miscellaneous Registers
        REG_CPURESET =          $302020 'Bit 2 - 0 :     Bit 0 for coprocessor engine,    Bit 1 for touch engine,    Bit 2 for audio engine.
        REG_PWM_DUTY =          $3020D4 'Bit 7 - 0 :  These bits define the backlight PWM output duty cycle
        REG_PWM_HZ =            $3020D0 'Bit 13 - 0 :  These bits define the backlight PWM output frequency in HZ.
        REG_INT_MASK =          $3020B0 'Bit 7 - 0 :   These bits are used to mask the corresponding interrupt.            
        REG_INT_EN =            $3020AC 'Bit 0 :  The host can set this bit to 1 to enable the global interrupt.
        REG_INT_FLAGS =         $3020A8 'Bit 7 - 0 :   These bits are interrupt flags set by the FT81X
        REG_GPIO_DIR =          $302090 'This register is a legacy register for backward comaptibility only
        REG_GPIO =              $302094 'This register is a legacy register for backward compatibility only
        REG_GPIOX_DIR =         $302098 'Bit 3-0: Controlling the direction of pin GPIO 3-0
        REG_GPIOX =             $30209C 'Bit 3-0: Writing or reading the pin of GPIO 3-0. 1 for high and 0 for low.
        REG_FREQUENCY =         $30200C 'Bit 31 - 0:  These bits are set 0x3938700 after reset, i.e.  The main clock frequency is 60MHz by default.
        REG_CLOCK =             $302008 'Bit 31 - 0:  These bits are set to zero after reset.   The register counts the number of FT81X main clock cycles since reset.
        REG_FRAMES =            $302004 'Bit 31 - 0:  These bits are set to zero after reset.   The register counts the number of screen frames.
        REG_ID =                $302000 'Bit 7 - 0:  These bits are the built-in register ID.  The  value shall always be 0x7C.
        REG_TRIM =              $10256C 'Bit 0 - 4:  These bits are set to trim the internal clock.
        REG_SPI_WIDTH =         $000180 'Bit 2: Extra dummy on SPI read transfer. Bit 1 - 0:   SPI data bus width

CON 'Addresses and Memory Map
        RAM_G =                 $000000  'General purpose graphics RAM, 1024 kB
        ROM_FONT =              $1E0000  'Font table and bitmap, 1152 kB
        ROM_FONT_ADDR =         $2FFFFC  'Font table pointer address, 4 B
        RAM_DL =                $300000  'Display List RAM, 8 kB
        RAM_REG =               $302000  'Registers, 4 kB 
        RAM_CMD =               $308000  'Command buffer, 4 kB       

        



PRI WrDl32(data) 'Write 32 bits to display list
  WriteReg32(DisplayListPointer,data)
  DisplayListPointer+=4


PRI WR16(address,data)
  WriteReg16(address,data)
PRI WR8(address,data)
  WriteReg8(address,data)
PRI WR32(address,data)
  WriteReg32(address,data)
PRI RD8(address)
  return ReadReg8(address)  

PRI WriteReg8(Address,data)  '8 bits
  address|=$800000
  low(pin_csn)
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 24, address)   'send address
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, data)   'send data
  high(pin_csn)

PRI WriteReg16(Address,data) '16 bits
  address|=$800000
  low(pin_csn)
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 24, address)   'send address
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, data&$FF)   'send data
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, (data>>8)&$FF)   'send data
  high(pin_csn)

PRI WriteReg32(Address,data) '32 bits
  address|=$800000
  low(pin_csn)
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 24, address)   'send address
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, data&$FF)   'send data
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, (data>>8)&$FF)   'send data
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, (data>>16)&$FF)   'send data    
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 8, (data>>24)&$FF)   'send data    
  high(pin_csn)    

PRI ReadReg8(Address)|data '8-bits
  low(pin_csn)
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 32, address<<8)   'send address + dummy byte
  data:=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8) 'read data
  high(pin_csn)
  return data
  
PRI ReadReg16(Address)|data '16 bits
  low(pin_csn)
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 32, address<<8)   'send address + dummy byte
  data:=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8) 'read data
  data+=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8)<<8 'read data  
  high(pin_csn)
  return data  

PRI ReadReg32(Address)|data '32 bits
  low(pin_csn)
  SPI.SHIFTOUT(pin_mosi, pin_sck, SPI#MSBFIRST , 32, address<<8)   'send address + dummy byte
  data:=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8) 'read data
  data+=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8)<<8 'read data  
  data+=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8)<<16 'read data  
  data+=SPI.SHIFTIN(pin_miso, pin_sck, SPI#MSBPRE, 8)<<24 'read data  
  high(pin_csn)
  return data

PUB HIGH(Pin)
    dira[Pin]~~
    outa[Pin]~~
         
PUB LOW(Pin)
    dira[Pin]~~
    outa[Pin]~  
      